import pyvista as pv
import xarray as xr
import os
import xml.etree.ElementTree as ET
import numpy as np
import pandas as pd
import dask
import dask.array as da
import re
import ESMF
import pathlib
from rasterio.warp import transform
import rioxarray # for xarray.rio

def _get_shape(mesh, dxdy):
    # number of cells @ this dxdy
    # numX = int((Emesh.coords[nodes][u].max() - Emesh.coords[nodes][u].min()) / dxdy)
    # numY = int((Emesh.coords[nodes][v].max() - Emesh.coords[nodes][v].min()) / dxdy)

    x = np.abs(mesh.bounds[0] - mesh.bounds[1])
    y = np.abs(mesh.bounds[2] - mesh.bounds[3])

    return int(x/dxdy),int(y/dxdy)


@dask.delayed
def _load_vtu(fname):
    # print('actually loading vtu ' + fname[0])
    vtu = pv.MultiBlock([pv.read(f) for f in fname])
    vtu = vtu.combine()
    return vtu


@dask.delayed
def _regrid_mesh_to_grid(vtu, dxdy, var):

    # print('regrid triggered')
    mesh, grid = _build_regridding_ds(vtu, dxdy)

    srcfield = ESMF.Field(mesh, name=var, meshloc=ESMF.MeshLoc.ELEMENT)
    srcfield.data[:] = vtu[var]
    dstfield = ESMF.Field(grid, var)
    dstfield.data[...] = np.nan
    # Do not use the 2nd order Conserve!
    # Another difference is that the second-order method does not guarantee that after regridding the range of values
    # in the destination field is within the range of values in the source field. For example, if the mininum value
    # in the source field is 0.0, then it's possible that after regridding with the second-order method, the
    # destination field will contain values less than 0.0.
    # https://esmf-org.github.io/dev_docs/ESMF_refdoc/node3.html#SECTION03023000000000000000
    regrid = ESMF.Regrid(srcfield, dstfield, regrid_method=ESMF.RegridMethod.CONSERVE, unmapped_action=ESMF.UnmappedAction.IGNORE)
    out = regrid(srcfield, dstfield, zero_region=ESMF.Region.SELECT)

    df = da.from_array(out.data)

    return df

# @dask.delayed
def _build_regridding_ds(mesh, dxdy):

    # mesh = pv.MultiBlock([pv.read(f) for f in fname])
    # mesh = mesh.combine()

    nodes, elements = (0, 1)
    u, v = (0, 1)

    Emesh = ESMF.Mesh(parametric_dim=2, spatial_dim=2)
    num_node = mesh.n_points
    num_elem = mesh.n_cells

    nodeId = np.linspace(0, num_node, num_node)
    nodeCoord = np.empty(num_node * 2)  # bc xy coords
    nodeOwner = np.zeros(num_node)

    for i in range(0, num_node * 2, 2):
        # just x, y
        nodeCoord[i] = mesh.GetPoint(i // 2)[0]
        nodeCoord[i + 1] = mesh.GetPoint(i // 2)[1]

    Emesh.add_nodes(num_node, nodeId, nodeCoord, nodeOwner)

    elemId = np.empty(num_elem)
    elemType = np.empty(num_elem)
    elemConn = np.empty(num_elem * 3)
    elemCoord = np.empty(num_elem * 2)

    for i in range(0, num_elem * 3, 3):
        elemId[i // 3] = i

        elemType[i // 3] = ESMF.MeshElemType.TRI

        v0 = mesh.GetCell(i // 3).GetPointId(0)
        v1 = mesh.GetCell(i // 3).GetPointId(1)
        v2 = mesh.GetCell(i // 3).GetPointId(2)

        elemConn[i] = v0
        elemConn[i + 1] = v2
        elemConn[i + 2] = v1

        centreX = 1 / 3 * (nodeCoord[v0] + nodeCoord[v1] + nodeCoord[v2])
        centreY = 1 / 3 * (nodeCoord[v0 + 1] + nodeCoord[v1 + 1] + nodeCoord[v2 + 1])

        elemCoord[i // 3] = centreX
        elemCoord[(i // 3) + 1] = centreY

    Emesh.add_elements(num_elem, elemId, elemType, elemConn, element_coords=elemCoord)

    numX, numY = _get_shape(mesh, dxdy)
    # numX = int((Emesh.coords[nodes][u].max() - Emesh.coords[nodes][u].min()) / dxdy)
    # numY = int((Emesh.coords[nodes][v].max() - Emesh.coords[nodes][v].min()) / dxdy)


    # cell centres
    x_center = np.linspace(start=Emesh.coords[nodes][u].min() + dxdy / 2.,
                           stop=Emesh.coords[nodes][u].max() - dxdy / 2., num=numX)
    y_center = np.linspace(start=Emesh.coords[nodes][v].min() + dxdy / 2.,
                           stop=Emesh.coords[nodes][v].max() - dxdy / 2., num=numY)

    # node coords
    x_corner = np.linspace(start=Emesh.coords[nodes][u].min(), stop=Emesh.coords[nodes][u].max(), num=numX + 1)
    y_corner = np.linspace(start=Emesh.coords[nodes][v].min(), stop=Emesh.coords[nodes][v].max(), num=numY + 1)

    max_index = np.array([len(x_center), len(y_center)])

    grid = ESMF.Grid(max_index, staggerloc=[ESMF.StaggerLoc.CENTER, ESMF.StaggerLoc.CORNER],
                     coord_sys=ESMF.CoordSys.CART)

    # RLO: access Grid center coordinates
    gridXCenter = grid.get_coords(0)
    gridYCenter = grid.get_coords(1)

    # RLO: set Grid center coordinates as a 2D array (this can also be done 1d)
    gridXCenter[...] = x_center.reshape((x_center.size, 1))
    gridYCenter[...] = y_center.reshape((1, y_center.size))

    # RLO: access Grid corner coordinates
    gridXCorner = grid.get_coords(0, staggerloc=ESMF.StaggerLoc.CORNER)
    gridYCorner = grid.get_coords(1, staggerloc=ESMF.StaggerLoc.CORNER)

    # RLO: set Grid corner coordinats as a 2D array
    gridXCorner[...] = x_corner.reshape((x_corner.size, 1))
    gridYCorner[...] = y_corner.reshape((1, y_corner.size))

    return Emesh, grid



def pvd_to_xarray(fname, dxdy=30, variables=None):
    """
    Opens a single vtu file or a pvd file and returns a delayed xarray object
    timestep basis
    :param fname: pvd file path
    :return:
    """

    # see if we were given a single vtu file or a pvd xml file
    filename, file_extension = os.path.splitext(fname)

    timesteps = None
    dt = np.timedelta64(1,'s')

    if file_extension != '.pvd':
        raise Exception('Not a pvd file')

    parse = ET.parse(fname)
    pvd = parse.findall(".//*[@file]")

    timesteps = parse.findall(".//*[@timestep]")

    epoch = np.datetime64(int(timesteps[0].get('timestep')), 's')

    # before we can come up with a possible dt, we need to see if we have multiple rank MPI outputs which will have 1 file
    # per MPI rank with a [basename]<timestamp>_<MPIRANK>.vtu format

    ranks = []

    # print(pvd)
    # loop through to detect how many mpi ranks were used
    for f in pvd:

        m = re.findall('.+_([0-9]+).vtu', f.get('file'))
        m = int(m[0])
        if m not in ranks:
            ranks.append(m)
        else:
            break

    print(f'MPI ranks = {len(ranks)} detected in vtu')
    nranks = len(ranks)

    if len(timesteps) > 1 and len(timesteps) > nranks:
        dt = np.datetime64(int(timesteps[nranks].get('timestep')),'s') - np.datetime64(int(timesteps[0].get('timestep')),'s')
        print('dt='+str(dt))
    vtu_paths = []

    base_dir = pathlib.Path(fname).parent.absolute()
    for i in range(0, len(pvd), nranks):
        current_vtu_paths = []  # holds 1 timestep's vtu paths
        for r in range(nranks):
            vtu = pvd[i + r]
            vtu_file = vtu.get('file')
            path = os.path.join(base_dir, vtu_file)
            current_vtu_paths.append(path)

        vtu_paths.append(current_vtu_paths)

    blocks = pv.MultiBlock([pv.read(f) for f in vtu_paths[0]])

    shape = _get_shape(blocks, dxdy)

    proj4 = blocks[0]["proj4"][0]
    print(f'proj4 = {proj4}')
    if variables is not None:
        # ensure we are not asking for an output that is a vector
        for v in variables:
            if len(blocks[0][v].shape) > 1:
                raise Exception('Cannot specify a variable that is a vector output shape = (n,3) ')
    else:
        variables = []
        blacklist = ['proj4', 'global_id']
        for v in blocks[0].array_names:
            if len(blocks[0][v].shape) == 1 and v not in blacklist: # don't add the vectors
                a = v.replace('[param] ','_param_').replace('/','') # sanitize the name, this should be fixed in CHM though
                variables.append(a)


    delayed_vtu = {}
    for var in variables:
        delayed_vtu[var] = []

    # cell centres
    x_center = np.linspace(start=blocks.bounds[0] + dxdy / 2.,
                           stop=blocks.bounds[1] - dxdy / 2., num=shape[0])
    y_center = np.linspace(start=blocks.bounds[2] + dxdy / 2.,
                           stop=blocks.bounds[3] - dxdy / 2., num=shape[1])


    blocks = None
    for v in vtu_paths:

        # print(v)
        vtu = _load_vtu(v)

        for var in variables:
            df = _regrid_mesh_to_grid(vtu, dxdy, var)
            d = da.from_delayed(df,
                                shape=shape,
                                dtype=np.dtype('float64'))

            tmp = xr.DataArray(d.T, name=var,
                               coords={'y': y_center,
                                       'x': x_center},
                               dims=['y', 'x'])

            delayed_vtu[var].append(tmp)

    end_time = np.datetime64(int(timesteps[-1].get('timestep')), 's')

    _dt = int(dt.astype("timedelta64[s]")/np.timedelta64(1, 's'))

    times = pd.date_range(start = epoch, end = end_time, freq= f'{_dt} s' , name="time")

    for var, arrays in delayed_vtu.items():
        delayed_vtu[var] = xr.concat(arrays, dim=times)
        delayed_vtu[var].chunk({'time':1})

    ds = xr.Dataset(data_vars=delayed_vtu)
    ds = ds.rio.set_crs(proj4)

    # Compute the lon/lat coordinates with rasterio.warp.transform
    ny, nx = len(ds['y']), len(ds['x'])
    x, y = np.meshgrid(ds['x'], ds['y'])
    # Rasterio works with 1D arrays
    lon, lat = transform(ds.rio.crs, {'init': 'EPSG:4326'}, x.flatten(), y.flatten())
    lon = np.asarray(lon).reshape((ny, nx))
    lat = np.asarray(lat).reshape((ny, nx))
    ds.coords['lon'] = (('y', 'x'), lon)
    ds.coords['lat'] = (('y', 'x'), lat)

    return ds






#!/usr/bin/env python

import numpy as np
import os 
import glob
import re
import sys

if len(sys.argv) == 1:
    print('Requires input path')
    exit(1)

searchpath = sys.argv[1]

files = glob.glob(f'{searchpath}/*.vrt')

if len(files) == 0:
    print('No vrt files found')
    exit(1)

# Find the vrt files
dirs = set()
suffix = None
basename = None
for f in files:
    m = re.findall(r'[0-9]+', f)
    m = m[-1]  #ensure we get the last number incase there are numbers in the filename
    dirs.add(int(m))
    
   
    s = re.findall(r'spd_up_(.+)\.vrt', f)
    
    if suffix is None:
        if len(s) > 0:
            suffix = s[-1]
        
    #figure out our base path
    if basename is None:
        m = re.findall(r'/(.+)_[0-9]+_[UV].vrt', f)
        if len(m) > 0:
            basename = m[-1]
        
dirs = list(dirs)
dirs = np.sort(dirs)

dtheta = dirs[1]-dirs[0]

ncat = len(dirs) # number of wind maps

print(f'Found {ncat} directions with step size = {dtheta} degrees')
print(f'Using basename = {basename}')

with open('config_WN.txt','w') as file:
    for i, ww in enumerate(dirs):

        if(i==0):
            i=ncat  

        for var in ['U','V','spd_up']:

            v = var
            if 'spd_up' in var:
                v = f'{var}_{suffix}'

            fname = f'{basename}_{ww}_{v}.vrt'

            p = os.path.abspath(searchpath + '/' + fname)

            NinjaName = f'Ninja{i}'

            if 'spd_up' not in var:
                NinjaName = f'{NinjaName}_{var}'

            line = "'%s' : {'file':'%s','method':'mean'}, \n" % (NinjaName,p)
            file.write(line)
            print(line)


import vtk
from vtk.util import numpy_support as VN

import xml.etree.ElementTree as ET
import numpy as np
import os


def mesh():

    vtu_files = []
    def __init__(self):
        pass

    def open_mesh(self, fname):
        """
        Opens a single vtu file or a pvd file. Only gets file handles. The actual read needs to be done on a per
        timestep basis
        :param self:
        :param fname: vtu or pvd file path
        :return:
        """

        # see if we were given a single vtu file or a pvd xml file
        filename, file_extension = os.path.splitext(fname)
        is_pvd = False
        pvd = [fname]  # if not given a pvd file, make this iterable for the below code
        timesteps = None
        dt = 1
        epoch = np.datetime64(0,'s')

        if file_extension == '.pvd':
            print('Detected pvd file, processing all linked vtu files')
            is_pvd = True
            parse = ET.parse(fname)
            pvd = parse.findall(".//*[@file]")

            timesteps = parse.findall(".//*[@timestep]")

            epoch = np.datetime64(int(timesteps[0].get('timestep')),'s')
            if len(timesteps) > 1:
                dt = int(timesteps[1].get('timestep')) - int(timesteps[0].get('timestep'))

        for vtu in pvd:
            path = vtu
            vtu_file = ''
            if is_pvd:
                vtu_file = vtu.get('file')
                path = fname[:fname.rfind('/') + 1] + vtu_file
            else:
                base = os.path.basename(path)  # since we have a full path to vtu, we need to get just the vtu filename
                vtu_file = os.path.splitext(base)[0]  # we strip out vtu later so keep here

            vtu_files.append(vtu_file)

    def read_timestep(self, timestep):
        if timestep > len(vtu_files):
            raise Exception("Index out of bounds")

        reader = vtk.vtkXMLUnstructuredGridReader()
        reader.SetFileName(vtu_files[timestep])

        mesh = reader.GetOutput()

        return mesh

    def get_all_triangle(self, idx):

        mesh = self.read_timestep(idx)
        # Get coords
        points = VN.vtk_to_numpy(mesh.GetPoints().GetData())

        # Get location of nodes
        X = [cpt[0] for cpt in points]
        Y = [cpt[1] for cpt in points]
        E = [cpt[2] for cpt in points]

        # Build face adjacency index
        triang = []
        for i in range(0, mesh.GetNumberOfCells()):
            v0 = mesh.GetCell(i).GetPointId(0)
            v1 = mesh.GetCell(i).GetPointId(1)
            v2 = mesh.GetCell(i).GetPointId(2)

            triang.append([v0, v1, v2])

        return {'X': X, 'Y': Y, 'E': E, 'triang': triang}

    def get_mesh_as_numpy(self, timestep, variable):

        mesh = self.read_timestep(timestep)
        cvar = VN.vtk_to_numpy(mesh.GetCellData().GetArray(variable))

        return cvar


#!/usr/bin/env python

import vtk
from osgeo import gdal, ogr, osr
from osgeo.gdalconst import GA_ReadOnly
import math
import imp
import sys
import xml.etree.ElementTree as ET
import subprocess
import numpy as np
import os
import re

import resource

import xarray as xr

gdal.UseExceptions()  # Enable errors

# Print iterations progress
# http://stackoverflow.com/a/34325723/410074
def printProgress(iteration, total, prefix='', suffix='', decimals=2, barLength=100):
    """
    Call in a loop to create terminal progress bar
    @params:
        iterations  - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
    """
    filledLength = int(round(barLength * iteration / float(total)))
    percents = round(100.00 * (iteration / float(total)), decimals)
    bar = '#' * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('%s [%s] %s%s %s\r' % (prefix, bar, percents, '%', suffix)),
    sys.stdout.flush()
    if iteration == total:
        print("\n")

import contextlib


#suppress stderr from one of the vtk calls that has deprecated warning out put for something outside of our control with vtk8.1
#https://stackoverflow.com/a/17753620/410074
@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:


    with stdchannel_redirected(sys.stderr, os.devnull):
        if compiler.has_function('clock_gettime', libraries=['rt']):
            libraries.append('rt')
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, 'w')
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()

# layer Input USM to be rasterized
# srsout Output CRS
# the filename of the raster
# pixel size, in srsout units. Constant dx,dy
# attribute The value from the USM we want to burn into the raster
def rasterize(layer, srsout, target_fname, pixel_size, attribute, user_define_extent=False, o_xmin=None, o_ymin=None, o_xmax=None, o_ymax=None, all_touched=True):
    x_min, x_max, y_min, y_max = layer.GetExtent()

    NoData_value = -9999
    # width/height of created raster in pixels.
    xsize = int((x_max - x_min) / pixel_size)
    ysize = int((y_max - y_min) / pixel_size)


    target_ds = gdal.GetDriverByName('GTiff').Create(target_fname, xsize, ysize, 1, gdal.GDT_Float32)
    target_ds.SetProjection(srsout.ExportToWkt())
    target_ds.SetGeoTransform((x_min, pixel_size, 0, y_max, 0, -pixel_size))
    band = target_ds.GetRasterBand(1)
    band.SetNoDataValue(NoData_value)

    # Rasterize. If all_touched is false, then only the cell centres that are within the triangle are used
    # if lower resolution rasterization is used, then this can miss cells. When all touched is true, as long as a triangle
    # touches a cell, then it's counted
    if all_touched:
        gdal.RasterizeLayer(target_ds, [1], layer, burn_values=[0],
                            options=['ALL_TOUCHED=TRUE',"ATTRIBUTE=" + attribute])
    else:
        gdal.RasterizeLayer(target_ds, [1], layer, burn_values=[0], options=[ "ATTRIBUTE=" + attribute]) ,

    target_ds = None

    # Optional clip file
    if user_define_extent: # or constrain_flag:
        subprocess.check_call(
            ['gdalwarp -overwrite -s_srs \"%s\" -t_srs \"%s\" -te %s %s %s %s -r \"%s\" -tr %s %s \"%s\" \"%s\"' %
             (srsout.ExportToProj4(),
              srsout.ExportToProj4(),
              o_xmin, o_ymin, o_xmax, o_ymax,
              'average', #var_resample_method[var]
              pixel_size, pixel_size,
              target_fname,
              target_fname[:-4] + '_clipped.tif')], shell=True)

def main():
    gdal.UseExceptions()  # Enable errors

    #####  load user configurable paramters here    #######
    # Check user defined configuraiton file
    if len(sys.argv) == 1:
        print('ERROR: main.py requires one argument [configuration file] (i.e. python main.py vtu2geo_config.py)')
        return

    # Get name of configuration file/module
    configfile = sys.argv[1]

    # Load in configuration file as module
    X = imp.load_source('',configfile)


    # if a 2nd command line argument is present, it is the input_path so use that, otherwise try to use the one from passed script
    input_path = ''
    if len(sys.argv) == 3: # we have a 2nd CLI arg
        input_path = sys.argv[2]
        if hasattr(X,'input_path'):
            print('Warning: Overwriting script defined input path with CL path')
    elif hasattr(X,'input_path'):
        input_path = X.input_path
    else:
        print('ERROR: No input path. A pvd or vtu file must be specified.')
        exit(-1)

    if os.path.isdir(input_path):
        print('ERROR: Either a pvd or vtu file must be specified.')
        exit(-1)

    variables = X.variables

    parameters = []
    if hasattr(X,'parameters'):
        parameters = X.parameters

    # Check if we want to constrain output to a example geotif
    constrain_flag = False
    if hasattr(X,'constrain_tif_file'):
        constrain_tif_file = X.constrain_tif_file
        var_resample_method = X.var_resample_method
        param_resample_method = X.param_resample_method
        constrain_flag = True

    output_path = input_path[:input_path.rfind('/')+1]

    if hasattr(X,'output_path'):
        output_path = X.output_path

        if not os.path.isdir(output_path):
            os.makedirs(output_path)

    pixel_size = 10
    if hasattr(X,'pixel_size'):
        pixel_size = X.pixel_size
    else:
        print('Default pixel size of 10 mx10 m will be used.')

    user_define_extent = False
    if hasattr(X,'user_define_extent'):
        user_define_extent = X.user_define_extent

    #Produces a lat/long regular grid in CF netCDF format instead of the TIF files
    nc_archive = False
    if hasattr(X,'nc_archive'):
        nc_archive = X.nc_archive

    #user defined output EPSG to use instead of the proj4 as defined in the vtu
    out_EPSG = None
    if hasattr(X,'out_EPSG'):
        out_EPSG = X.out_EPSG

    if parameters is not None and nc_archive:
        print('Parameters are ignored when writing the nc archive.')
        parameters = []

    all_touched = True
    if hasattr(X,'all_touched'):
        all_touched = X.all_touched

    keep_tifs = False
    if hasattr(X,'keep_tifs'):
        keep_tifs = X.keep_tifs

    #####
    reader = vtk.vtkXMLUnstructuredGridReader()

    # see if we were given a single vtu file or a pvd xml file
    filename, file_extension = os.path.splitext(input_path)
    is_pvd = False
    pvd = [input_path] # if not given a pvd file, make this iterable for the below code
    timesteps=None
    if file_extension == '.pvd':
        print('Detected pvd file, processing all linked vtu files')
        is_pvd = True
        if not os.path.exists(input_path):
            print(f'ERROR: Path does not exists: {input_path}')
            exit(-1)

        parse = ET.parse(input_path)
        pvd = parse.findall(".//*[@file]")

        timesteps = parse.findall(".//*[@timestep]")

    o_xmin = o_xmax = o_ymin = o_ymax = None
    if user_define_extent:
        o_xmin = X.o_xmin
        o_xmax = X.o_xmax
        o_ymin = X.o_ymin
        o_ymax = X.o_ymax

    # Get info for constrained output extent/resolution if selected
    if constrain_flag :
        ex_ds = gdal.Open(constrain_tif_file,GA_ReadOnly)
        gt = ex_ds.GetGeoTransform()
        pixel_width = np.abs(gt[1])
        pixel_height = np.abs(gt[5])
        # Take extent from user input
        if user_define_extent:
            o_xmin = X.o_xmin
            o_xmax = X.o_xmax
            o_ymin = X.o_ymin
            o_ymax = X.o_ymax
        else: # Get extent for clipping from input tif
            o_xmin = gt[0]
            o_ymax = gt[3]
            o_xmax = o_xmin + gt[1] * ex_ds.RasterXSize
            o_ymin = o_ymax + gt[5] * ex_ds.RasterYSize

        print(("Output pixel size is " + str(pixel_width) + " by " + str(pixel_height)))
        ex_ds = None

    if constrain_flag:
        print(" Constrain flag currently not supported!")
        return -1


    files_processed=1 # this really should be 1 for useful output

    #information to build up the nc meta data
    nc_rasters = {}
    if nc_archive:
        for v in variables:
            nc_rasters[v]=[]

    nc_time_counter=0
    tifs_to_remove = []
    epoch=np.datetime64(0,'s')

    # if we are loading a pvd, we have access to the timestep information if we want to build
    if is_pvd and timesteps is not None:
        epoch = np.datetime64(int(timesteps[0].get('timestep')),'s')

    dt = 1 # model timestep, in seconds

    # before we can come up with a possible dt, we need to see if we have the MPI outputs which will have 1 file
    # per MPI rank with a [basename]<timestamp>_<MPIRANK>.vut format

    is_mpi_output = False
    ranks = []
    nranks = 0
    m = re.findall('.+_([0-9]+).vtu', pvd[0])
    if len(m) > 0 :
        is_mpi_output = True
        # loop through to detect how many mpi ranks were used

        for f in pvd:
            m = re.findall('.+_([0-9]+).vtu', f)
            m = int(m[0])
            if m not in ranks:
                ranks.append(m[0])
            else:
                break

        print(f'MPI ranks = {len(ranks)} detected in vtu')
        nranks = len(ranks) - 1 # 0-index!!

    # reorganize the flat structure to handle MPI
    for file, ts in zip(pvd, timesteps):

    if timesteps is not None and len(timesteps) > 1:
        if not is_mpi_output:
            dt = int(timesteps[1].get('timestep')) - int(timesteps[0].get('timestep'))
        else:
            if len(timesteps) > len(ranks):
                dt = int(timesteps[nranks + 1].get('timestep')) - int(timesteps[0].get('timestep'))

    print(('Start epoch: %s, model dt = %i (s)' %(epoch,dt)))

    # because of how the netcdf is built we hold a file of file handles before converting. ensure we can do so
    if nc_archive:
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        total_output_files = len(pvd) * (len(variables)+len(parameters))
        if soft < total_output_files or hard < total_output_files:
          
            print('The system soft or hard file limit is less than the total number of tmp files to be created.')
            print('The system ulimit will be raised to at least ' + str(total_output_files))

            try:
                resource.setrlimit(resource.RLIMIT_NOFILE, (total_output_files,resource.RLIM_INFINITY ) )
            except ValueError as e:
                print("Failed to raise the limit")
                raise e


    # If we have MPI output we need this counter to keep track of what MPI rank we are currently on
    current_MPI_rank = 0
    current_MPI_rank_tiffs = []

    for vtu in pvd:
        path = vtu
        vtu_file = ''
        if is_pvd:
            vtu_file = vtu.get('file')
            path = input_path[:input_path.rfind('/')+1] + vtu_file
        else:
            base = os.path.basename(path) # since we have a full path to vtu, we need to get just the vtu filename
            vtu_file = os.path.splitext(base)[0] #we strip out vtu later so keep here

        if not os.path.exists(path):
            print(f'ERROR: Path does not exists: {path}')
            exit(-1)

        printProgress(files_processed,len(pvd),decimals=0)
        reader.SetFileName(path)
        reader.Update()
        #shut up a deprecated warning from vtk 8.1
        # with stdchannel_redirected(sys.stderr, os.devnull):
        #     reader.Update()

        mesh = reader.GetOutput()

        #default the pixel size to (min+max)/2
        if not pixel_size:
            area_range = mesh.GetCellData().GetArray('Area').GetRange()
            pixel_size = (math.sqrt(area_range[0]) + math.sqrt(area_range[1]))/2
            pixel_size = int( math.ceil(pixel_size) )

        driver = ogr.GetDriverByName('Memory')
        output_usm = driver.CreateDataSource('out')

        srsin = osr.SpatialReference()

        if not mesh.GetFieldData().HasArray("proj4"):
            print("VTU file does not contain a proj4 field")
            return -1

        vtu_proj4 = mesh.GetFieldData().GetAbstractArray("proj4").GetValue(0)
        srsin.ImportFromProj4(vtu_proj4)

        is_geographic = srsin.IsGeographic()

        #output conic equal area for geotiff
        srsout = osr.SpatialReference()
        srsout.ImportFromProj4(vtu_proj4)

        if out_EPSG:
            srsout.ImportFromEPSG(out_EPSG)

        trans = osr.CoordinateTransformation(srsin,srsout)
        layer = output_usm.CreateLayer('poly', srsout, ogr.wkbPolygon)
        cd = mesh.GetCellData()

        for i in range(0,cd.GetNumberOfArrays()):
            layer.CreateField(ogr.FieldDefn(cd.GetArrayName(i), ogr.OFTReal))

        #build the triangulation geometery
        for i in range(0, mesh.GetNumberOfCells()):
            v0 = mesh.GetCell(i).GetPointId(0)
            v1 = mesh.GetCell(i).GetPointId(1)
            v2 = mesh.GetCell(i).GetPointId(2)

            ring = ogr.Geometry(ogr.wkbLinearRing)
            if is_geographic:
                scale = 100000.  #undo the scaling that CHM does for the paraview output
                ring.AddPoint( mesh.GetPoint(v0)[0] / scale, mesh.GetPoint(v0)[1] / scale)
                ring.AddPoint (mesh.GetPoint(v1)[0]/ scale, mesh.GetPoint(v1)[1]/ scale )
                ring.AddPoint( mesh.GetPoint(v2)[0]/ scale, mesh.GetPoint(v2)[1]/ scale )
                ring.AddPoint( mesh.GetPoint(v0)[0]/ scale, mesh.GetPoint(v0)[1]/ scale ) # add again to complete the ring.
            else:
                ring.AddPoint( mesh.GetPoint(v0)[0], mesh.GetPoint(v0)[1] )
                ring.AddPoint (mesh.GetPoint(v1)[0], mesh.GetPoint(v1)[1] )
                ring.AddPoint( mesh.GetPoint(v2)[0], mesh.GetPoint(v2)[1] )
                ring.AddPoint( mesh.GetPoint(v0)[0], mesh.GetPoint(v0)[1] ) # add again to complete the ring.

            ring.Transform(trans)
            tpoly = ogr.Geometry(ogr.wkbPolygon)
            tpoly.AddGeometry(ring)

            feature = ogr.Feature(layer.GetLayerDefn())
            feature.SetGeometry(tpoly)

            if variables is None:
                variables = []
                for j in range(0, cd.GetNumberOfArrays()):
                    name = cd.GetArrayName(j)
                    if '[param]' not in name:
                        variables.append(name)

            for v in variables:
                try:
                    data = cd.GetArray(v).GetTuple(i)
                    feature.SetField(str(v), float(data[0]))
                except:
                    print(("Variable %s not present in mesh" % v))
                    return -1

            if parameters is not None:
                for p in parameters:
                    try:
                        data = cd.GetArray(p).GetTuple(i)
                        feature.SetField(str(p), float(data[0]))
                    except:
                        print(("Parameter %s not present in mesh" % v))
                        return -1
            layer.CreateFeature(feature)

        for var in variables:
            print(var)
            target_fname = os.path.join(output_path,
                                        vtu_file + '_' + var.replace(" ", "_") + str(pixel_size) + 'x' + str(
                                            pixel_size) + '.tif')
            rasterize(layer, srsout, target_fname, pixel_size, var, user_define_extent, o_xmin, o_ymin, o_xmax, o_ymax, all_touched)

            if nc_archive:
                df = xr.open_rasterio(target_fname).sel(band=1).drop('band')
                df = df.rename({'x':'lon','y':'lat'})
                df.coords['time']=epoch + nc_time_counter*np.timedelta64(dt,'s') # this will automatically get converted to min or hours in the output nc
                df.name=var
                nc_rasters[var].append(df) # these are lazy loaded at the to_netcdf call

                # remove the tifs we produce
                tifs_to_remove.append(target_fname)


        if parameters is not None:
            print(parameters)
            for p in parameters:
                target_param_fname = os.path.join(output_path, vtu_file + '_'+ p.replace(" ","_") + str(pixel_size)+'x'+str(pixel_size)+'.tif')
                rasterize(layer, srsout, target_param_fname, pixel_size, p, all_touched = all_touched)

        nc_time_counter += 1
        # we don't need to dump parameters for each timestep as they are currently assumed invariant with time.
        parameters = None

        #no parameters and no variables, just exit at this point
        if not variables and parameters is None:
            break

        files_processed += 1

    if nc_archive:

        datasets = []

        for var, rasters in list(nc_rasters.items()):
            a = xr.concat(rasters,dim='time')
            datasets.append(a.to_dataset())

        arr = xr.merge(datasets)
        print('Writing netCDF file')
        fname = os.path.join(os.path.splitext(input_path)[0]+'.nc')
        arr.to_netcdf(fname,engine='netcdf4')

        for f in tifs_to_remove:
            try:
                os.remove(f)
            except:
                pass

if __name__ == "__main__":

    main()



import xarray as xr
import glob
import os
import pandas as pd
import datetime


def pts_dir_to_nc(path, output_nc, UTC_offset=0):
    """
    Converts a out put points directory to a netcdf file
    :param path:
    :param output_nc:
    :param UTC_offset:
    :return:
    """
    files = glob.glob( os.path.join(path,'*_out.txt') )

    pts_to_nc(files, output_nc, UTC_offset)


def pts_dir_to_pd(path,  UTC_offset=0):
    """
    Converts an output points directory to a Pandas dataframe
    :param path:
    :param UTC_offset:
    :return:
    """
    files = glob.glob( os.path.join(path,'*_out.txt') )

    return pts_to_pd(files, UTC_offset)


def pts_to_pd(files, UTC_offset = 0):
    """
    Loads output point files into one large Pandas data frame
    :param files: List of input files
    :param output_nc:
    :param UTC_offset:
    :return:
    """
    # Time zone
    UTC_to_MST = 0  # UTC

    # Loop each station
    CHM_pts = []
    for cf in files:

        cSta = os.path.basename(cf).split('_')[0]
        print("processing " + cSta)

        # Importto pandas dataframe
        modData = pd.read_csv(cf, sep=",", parse_dates=True)
        modData.set_index('datetime', inplace=True)
        # Make datetime the index
        modData.index = pd.to_datetime(modData.index)
        modData.index = modData.index + datetime.timedelta(hours=UTC_offset)
        # Convert to data set and add station name
        c_ds = xr.Dataset(modData)
        c_ds['station'] = cSta
        c_ds = c_ds.rename({'datetime': 'time'})
        # Save in list
        CHM_pts.append(c_ds)

    # concat all stations
    ds = xr.concat(CHM_pts, dim='station')

    # Set -9999 (CHM missing value) to nan
    ds = ds.where(ds != -9999)

    if 'dm/dt' in ds.data_vars:
        ds = ds.drop(['dm/dt'])

    return ds


def pts_to_nc(files, output_nc, UTC_offset = 0):
    """
    Converts the point file to a netcdf file
    :param files: List of files
    :param output_nc: Output netcdf file
    :param UTC_offset: Conversion of UTC to desired timeframe
    :return:
    """

    ds = pts_to_pd(files,UTC_offset)

    # Save out to netcdf
    ds.to_netcdf(output_nc, engine='netcdf4')


from setuptools import setup
import subprocess
from setuptools import find_packages

def get_installed_gdal_version():
    try:
        version = subprocess.run(["gdal-config","--version"], stdout=subprocess.PIPE).stdout.decode()

        version = version.replace('\n', '')
        version = "=="+version+".*"
        return version
    except FileNotFoundError as e:
        raise(""" ERROR: Could not find the system install of GDAL. 
                  Please install it via your package manage of choice.
                """
            )

setup(
    name='CHM',
    version='1.0.1',
    description='python tools for CHM',
    license='MIT',
    packages=find_packages(),
    author='Chris Marsh',
    author_email='chris.marsh@usask.ca',
    install_requires=['vtk','numpy','xarray','netCDF4','pandas',
                      'pygdal'+get_installed_gdal_version(), 'dask[complete]',
                      'pyvista', 'pyESMF==8.1.0.2b36','rioxarray','pandas','rasterio'],
    scripts=["scripts/vtu2geo.py","scripts/windmapper2mesher.py"],
)